__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2020-2021, Vanessa Sochat"
__license__ = "MPL 2.0"

from django import forms


class TokenForm(forms.Form):
    token = forms.CharField(label="Notebook token", max_length=100)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2020-2021, Vanessa Sochat"
__license__ = "MPL 2.0"

from django.contrib.auth.admin import UserAdmin
from django.contrib import admin
from snakeface.apps.users.models import User

admin.site.register(User, UserAdmin)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2020-2021, Vanessa Sochat"
__license__ = "MPL 2.0"

__version__ = "0.0.14"
AUTHOR = "Vanessa Sochat"
AUTHOR_EMAIL = "vsochat@stanford.edu"
NAME = "snakeface"
PACKAGE_URL = "https://github.com/snakemake/snakeface"
KEYWORDS = "snakemake,workflow management,pipeline,interface, workflows"
DESCRIPTION = "Snakemake Interface"
LICENSE = "LICENSE"

################################################################################
# Global requirements


INSTALL_REQUIRES = (
    ("snakedeploy", {"min_version": None}),
    ("snakemake", {"min_version": None}),
    ("pyaml", {"min_version": "20.4.0"}),
    ("Jinja2", {"min_version": "2.11.2"}),
    ("Django", {"exact_version": "3.0.8"}),
    ("django-q", {"min_version": "1.3.4"}),
    ("django-crispy-forms", {"min_version": "1.10.0"}),
    ("django-taggit", {"min_version": "1.3.0"}),
    ("django-gravatar", {"min_version": None}),
    ("django-ratelimit", {"min_version": "3.0.0"}),
    ("django-extensions", {"min_version": "3.0.2"}),
    ("djangorestframework", {"exact_version": "3.11.1"}),
    ("drf-yasg", {"exact_version": "1.20.0"}),
    ("channels", {"exact_version": "3.0.3"}),
    ("social-auth-app-django", {"min_version": "4.0.0"}),
    ("social-auth-core", {"min_version": "3.3.3"}),
)

# Dependencies provided by snakemake: pyYaml, jinja2

EMAIL_REQUIRES = (("sendgrid", {"min_version": "6.4.3"}),)
POSTGRES_REQUIRES = (("psycopg2-binary", {"min_version": "2.8.5"}),)

TESTS_REQUIRES = (("pytest", {"min_version": "4.6.2"}),)

ALL_REQUIRES = INSTALL_REQUIRES + EMAIL_REQUIRES + POSTGRES_REQUIRES



#!/usr/bin/env python

__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2020-2021, Vanessa Sochat"
__license__ = "MPL 2.0"

from django.contrib.auth import REDIRECT_FIELD_NAME
from django.shortcuts import resolve_url, redirect
from snakeface.settings import cfg
from snakeface import settings
from urllib.parse import urlparse


def login_is_required(
    function=None, login_url=None, redirect_field_name=REDIRECT_FIELD_NAME
):
    """
    Decorator to extend login required to also check if a notebook auth is
    desired first.
    """

    def wrap(request, *args, **kwargs):

        # If we are using a notebook, the user is required to login with a token
        if (cfg.NOTEBOOK or cfg.NOTEBOOK_ONLY) and not request.user.is_authenticated:
            return redirect("users:notebook_login")

        # If we have the token in the session
        elif (
            cfg.NOTEBOOK
            or cfg.NOTEBOOK_ONLY
            and request.session.get("notebook_auth")
            == request.session.get("notebook_token")
        ):
            return function(request, *args, **kwargs)

        # If the user is authenticated, return the view right away
        elif request.user.is_authenticated:
            return function(request, *args, **kwargs)

        # Otherwise, prepare login url (from django user_passes_test)
        # https://github.com/django/django/blob/master/django/contrib/auth/decorators.py#L10
        path = request.build_absolute_uri()
        resolved_login_url = resolve_url(login_url or settings.LOGIN_URL)
        login_scheme, login_netloc = urlparse(resolved_login_url)[:2]
        current_scheme, current_netloc = urlparse(path)[:2]
        if (not login_scheme or login_scheme == current_scheme) and (
            not login_netloc or login_netloc == current_netloc
        ):
            path = request.get_full_path()
        from django.contrib.auth.views import redirect_to_login

        return redirect_to_login(path, resolved_login_url, redirect_field_name)

    wrap.__doc__ = function.__doc__
    wrap.__name__ = function.__name__
    return wrap



from rest_framework.permissions import BasePermission, SAFE_METHODS
from rest_framework.authtoken.models import Token


class AllowAnyGet(BasePermission):
    """Allows an anonymous user access for GET requests only."""

    def has_permission(self, request, view):

        if request.user.is_anonymous and request.method == "GET":
            return (True,)

        if request.user.is_staff or request.user.is_superuser:
            return True

        return request.method in SAFE_METHODS


def check_user_authentication(request):
    """Given a request, check that the user is authenticated via a token in
    the header.
    """
    token = get_token(request)

    # Case 1: no token and auth is required, prompt the user for it
    if not token:
        return None, 401

    # Case 2: the token is not associated with a user
    try:
        token = Token.objects.get(key=token)
    except:
        return None, 403

    return token.user, 200


def get_token(request):
    """The same as validate_token, but return the token object to check the
    associated user.
    """
    # Coming from HTTP, look for authorization as bearer token
    token = request.META.get("HTTP_AUTHORIZATION")

    if token:
        token = token.split(" ")[-1].strip()
        try:
            return Token.objects.get(key=token)
        except Token.DoesNotExist:
            pass

    # Next attempt - try to get token via user session
    elif request.user.is_authenticated and not request.user.is_anonymous:
        try:
            return Token.objects.get(user=request.user)
        except Token.DoesNotExist:
            pass



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2020-2021, Vanessa Sochat"
__license__ = "MPL 2.0"

from django.conf import settings
from django.contrib.auth.models import AbstractUser, BaseUserManager
from django.db.models.signals import post_save
from django.dispatch import receiver

from django.db import models
from rest_framework.authtoken.models import Token


class CustomUserManager(BaseUserManager):
    def _create_user(
        self, username, email, password, is_staff, is_superuser, **extra_fields
    ):
        """
        Creates and saves a User with the given username, email and password.
        """
        if not username:
            raise ValueError("The given username must be set")

        email = self.normalize_email(email)
        user = self.model(
            username=username,
            email=email,
            is_staff=is_staff,
            is_active=True,
            **extra_fields
        )
        user.set_password(password)
        user.save(using=self._db)
        return user

    def create_user(self, username, email=None, password=None, **extra_fields):
        return self._create_user(
            username, email, password, False, False, **extra_fields
        )

    def create_superuser(self, username, email, password, **extra_fields):
        return self._create_user(username, email, password, True, True, **extra_fields)

    def add_superuser(self, user):
        """ Intended for existing user"""
        user.is_superuser = True
        user.save(using=self._db)
        return user

    def add_staff(self, user):
        """ Intended for existing user"""
        user.is_staff = True
        user.save(using=self._db)
        return user


class User(AbstractUser):
    active = models.BooleanField(default=True)

    # has the user agreed to terms?
    agree_terms = models.BooleanField(default=False)
    agree_terms_date = models.DateTimeField(blank=True, default=None, null=True)

    # Notebook token (only for terminal notebook)
    notebook_token = models.CharField(
        max_length=36, default=None, null=True, blank=True
    )

    # Ensure that we can add staff / superuser and retain on logout
    objects = CustomUserManager()

    class Meta:
        app_label = "users"

    @property
    def token(self):
        """The user token is for interaction with creating and updating workflows"""
        return str(Token.objects.get(user=self))

    def has_create_permission(self):
        """has create permission determines if the user (globally) can create
        new collections. By default, superusers and admin can, along with
        regular users if USER_COLLECTIONS is True. Otherwise, not.
        """
        if self.is_superuser is True or self.is_staff is True:
            return True
        if settings.USER_COLLECTIONS is True:
            return True
        return False

    def get_credentials(self, provider):
        """ return one or more credentials, or None"""
        if self.is_anonymous is False:
            try:
                # Case 1: one credential
                credential = self.social_auth.get(provider=provider)
                return credential
            except:
                # Case 2: more than one credential for the provider
                credential = self.social_auth.filter(provider=provider)
                if credential:
                    return credential.last()

    def get_providers(self):
        """return a list of providers that the user has credentials for."""
        return [x.provider for x in self.social_auth.all()]

    def get_label(self):
        return "users"


@receiver(post_save, sender=settings.AUTH_USER_MODEL)
def create_auth_token(sender, instance=None, created=False, **kwargs):
    """Create a token for the user when the user is created (with oAuth2)

    1. Assign user a token
    2. Assign user to default group

    Create a Profile instance for all newly created User instances. We only
    run on user creation to avoid having to check for existence on each call
    to User.save.

    """
    # This auth token is intended for APIs
    if created and "api" in settings.PLUGINS_ENABLED:
        Token.objects.create(user=instance)



import os

from channels.auth import AuthMiddlewareStack
from channels.routing import ProtocolTypeRouter, URLRouter
from django.core.asgi import get_asgi_application
from snakeface.apps.main import routing

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "snakeface.settings")

application = ProtocolTypeRouter(
    {
        "http": get_asgi_application(),
        "websocket": AuthMiddlewareStack(URLRouter(routing.websocket_urlpatterns)),
    }
)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2020-2021, Vanessa Sochat"
__license__ = "MPL 2.0"

from snakeface.settings import cfg
import subprocess
import threading

import tempfile
import os
import re


def get_workdir_choices(path=None):
    """Given the working directory set on init, return potential subdirectories."""
    path = path or cfg.WORKDIR
    choices = [(path, "/")]

    # Recursive to working directory is default
    for root, dirs, files in sorted(os.walk(path)):
        for f in sorted(dirs):
            if f == "__pycache__":
                continue
            fullpath = os.path.join(root, f)
            # Ignore all hidden files and paths
            if "/." in f or "/." in fullpath or "/." in root:
                continue
            choices.append((fullpath, fullpath))
    return choices


def get_snakefile_choices(path=None):
    """Given the working directory set on init, return all discovered snakefiles."""
    path = path or cfg.WORKDIR
    choices = []

    # Recursive to working directory is default
    for root, dirs, files in sorted(os.walk(path)):
        for f in sorted(files):
            fullpath = os.path.join(root, f)
            if re.search("Snakefile", f):
                choices.append((fullpath, fullpath))
    return choices


def write_file(filename, content):
    """Write some text content to a file"""
    with open(filename, "w") as fd:
        fd.write(content)
    return filename


def read_file(filename):
    """Write some text content to a file"""
    with open(filename, "r") as fd:
        content = fd.read()
    return content


def get_tmpfile(prefix="", suffix=""):
    """get a temporary file with an optional prefix. By default, the file
    is closed (and just a name returned).

    Arguments:
     - prefix (str) : prefix with this string
    """
    tmpdir = tempfile.gettempdir()
    prefix = os.path.join(tmpdir, os.path.basename(prefix))
    fd, tmp_file = tempfile.mkstemp(prefix=prefix, suffix=suffix)
    os.close(fd)
    return tmp_file


class ThreadRunner(threading.Thread):
    """We need to be able to run a Snakemake job as a thread, and kill it if
    an exception is raised based on it's id
    """

    def set_workflow(self, workflow):
        self.workflow = workflow

    @property
    def thread_id(self):
        """Return the id of the thread, either attributed to the class or
        by matching the Thread instance
        """
        if hasattr(self, "_thread_id"):
            return self._thread_id
        for thread_id, thread in threading._active.items():
            if thread is self:
                return thread_id


class CommandRunner(object):
    """Wrapper to use subprocess to run a command. This is based off of pypi
    vendor distlib SubprocesMixin.
    """

    def __init__(self):
        self.reset()

    def reset(self):
        self.error = []
        self.output = []
        self.retval = None

    def reader(self, stream, context):
        """Get output and error lines and save to command runner."""
        # Make sure we save to the correct field
        lines = self.error
        if context == "stdout":
            lines = self.output

        while True:
            s = stream.readline()
            if not s:
                break
            lines.append(s.decode("utf-8"))
        stream.close()

    def run_command(
        self, cmd, env=None, cancel_func=None, cancel_func_kwargs=None, **kwargs
    ):
        self.reset()
        cancel_func_kwargs = cancel_func_kwargs or {}

        # If we need to update the environment
        # **IMPORTANT: this will include envars from host. Absolutely cannot
        # be any secrets (they should be defined in the app settings file)
        envars = os.environ.copy()
        if env:
            envars.update(env)

        p = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=envars, **kwargs
        )

        # Create threads for error and output
        t1 = threading.Thread(target=self.reader, args=(p.stdout, "stdout"))
        t1.start()
        t2 = threading.Thread(target=self.reader, args=(p.stderr, "stderr"))
        t2.start()

        # Continue running unless cancel function is called
        counter = 0
        while True:

            # Check on process for finished or cancelled
            if p.poll() != None:
                print("Return value found, stopping.")
                break

            # Check the cancel function every 100 loops
            elif (
                counter % 10000 == 0
                and cancel_func
                and cancel_func(**cancel_func_kwargs)
            ):
                print("Process is terminated")
                p.terminate()
                break
            counter += 1

        # p.wait()
        t1.join()
        t2.join()
        self.retval = p.returncode
        return self.output



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2020, Vanessa Sochat"
__license__ = "MPL 2.0"

import uuid
import os


def get_notebook_token(request, verbose=True):
    """If a notebook token isn't defined, generate it (and print to the console)
    The token is used to generate a user to log the user in.
    """
    # If the user has already logged in, return current token
    token = request.session.get("notebook_token")

    # The user session has ended, but has authenticated before
    user = get_notebook_user()
    if request.user.is_authenticated:
        token = request.user.notebook_token

    # Second attempt, see if user has logged in before and retrieve token
    elif user:
        token = user.notebook_token

    if not token:
        token = str(uuid.uuid4())
        user = get_or_create_notebook_user(token)
        request.session["notebook_token"] = token
        request.session.modified = True
    if verbose:
        print("Enter token: %s" % token)
    return token


def get_username():
    """get the username based on the effective uid. This is for a notebook
    execution, and doesn't add any additional security, but rather is used for
    personalization and being able to create an associated django user.
    """
    try:
        import pwd

        return pwd.getpwuid(os.getuid())[0]
    except:
        return "snakeface-user"


def get_notebook_user():
    """Get the notebook user, if they have logged in before."""
    from snakeface.apps.users.models import User
    from snakeface.settings import cfg

    try:
        return User.objects.get(username=cfg.USERNAME)
    except:
        return None


def get_or_create_notebook_user(token):
    """Get or create the notebook user. Imports are done in the function because
    Django startup (settings.py) uses these functions.
    """
    from snakeface.apps.users.models import User
    from snakeface.settings import cfg

    try:
        user = User.objects.get(username=cfg.USERNAME)
    except:
        user = User.objects.create_user(cfg.USERNAME, None, token)
        user.notebook_token = token
        user.save()
    return user



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2020-2021, Vanessa Sochat"
__license__ = "MPL 2.0"

from django.conf.urls import include, url
from django.contrib import admin
from django.views.generic.base import TemplateView

from snakeface.apps.base import urls as base_urls
from snakeface.apps.main import urls as main_urls
from snakeface.apps.users import urls as user_urls
from snakeface.apps.api import urls as api_urls

admin.site.site_header = "Snakeface Admin"
admin.site.site_title = "Snakeface Admin"
admin.site.index_title = "Snakeface Admin"

admin.autodiscover()

urlpatterns = [
    url(r"^admin/", admin.site.urls),
    url(r"^", include(base_urls, namespace="base")),
    url(r"^", include(api_urls, namespace="api")),
    url(r"^", include(main_urls, namespace="main")),
    url(r"^", include(user_urls, namespace="users")),
    url(
        r"^robots\.txt?/$",
        TemplateView.as_view(
            template_name="base/robots.txt", content_type="text/plain"
        ),
    ),
]



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2020, Vanessa Sochat"
__license__ = "MPL 2.0"


from django.contrib.auth import logout as auth_logout, authenticate, login
from django.contrib import messages
from django.shortcuts import render, redirect
from ratelimit.decorators import ratelimit
from snakeface.apps.users.forms import TokenForm
from snakeface.apps.users.decorators import login_is_required

from snakeface.apps.users.utils import get_notebook_token
from snakeface.settings import (
    VIEW_RATE_LIMIT as rl_rate,
    VIEW_RATE_LIMIT_BLOCK as rl_block,
    cfg,
)


@login_is_required
def logout(request):
    """log the user out, either from the notebook or traditional Django auth"""
    auth_logout(request)

    # Notebook: delete both tokens to ensure we generate a new one on logout
    if cfg.NOTEBOOK or cfg.NOTEBOOK_ONLY:
        return redirect("users:notebook_login")

    # A traditional Django authentication is here
    return redirect("/")


@ratelimit(key="ip", rate=rl_rate, block=rl_block)
def notebook_login(request):
    """Given the user doesn't have a token in the request session, ask for it."""
    # If they came to page directly, we need to generate the token
    valid_token = get_notebook_token(request)
    form = TokenForm()

    # If the user is submitting the form, validate it
    if request.method == "POST":
        form = TokenForm(request.POST)
        if form.is_valid():

            # If the form is valid, get/create the user and log in
            if form.cleaned_data["token"] == valid_token:
                user = authenticate(username=cfg.USERNAME, password=valid_token)
                if not user:
                    messages.warning(request, "That token is not valid.")
                else:
                    login(request, user)
                    return redirect("main:dashboard")
            else:
                messages.warning(request, "That token is not valid.")
        else:
            return render(request, "login/notebook.html", {"form": form})

    # If a user is already authenticated, redirect to dashboard
    if request.user.is_authenticated:
        return redirect("main:dashboard")

    # If the token isn't provided, they need to provide it
    return render(request, "login/notebook.html", {"form": form})



"""

Copyright (C) 2017-2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from django.core.management.base import BaseCommand, CommandError

from shub.apps.users.models import User
from shub.logger import bot


class Command(BaseCommand):
    """remove superuser will remove admin privs for a singularity
    registry. The super user is an admin that can build, delete,
    and manage images
    """

    def add_arguments(self, parser):
        # Positional arguments
        parser.add_argument("--username", dest="username", default=None, type=str)

    help = "Removes superuser priviledges for the registry."

    def handle(self, *args, **options):
        if options["username"] is None:
            raise CommandError("Please provide a username with --username")

        bot.debug("Username: %s" % options["username"])

        try:
            user = User.objects.get(username=options["username"])
        except User.DoesNotExist:
            raise CommandError("This username does not exist.")

        if user.is_superuser is False:
            raise CommandError("This user already is not a superuser.")
        else:
            user.is_superuser = False
            bot.debug("%s is no longer a superuser." % (user.username))
            user.save()



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2020-2021, Vanessa Sochat"
__license__ = "MPL 2.0"

from django.forms import ModelForm
from snakeface.apps.main.models import Workflow
from snakeface.apps.main.utils import get_workdir_choices
from django import forms
from snakeface.settings import cfg


class WorkflowForm(ModelForm):
    workdirs = forms.ChoiceField(choices=get_workdir_choices())

    class Meta:
        model = Workflow

        # Notebooks are always implicitly private, there is only one user
        if cfg.NOTEBOOK:
            fields = ["name", "workdirs"]
        else:
            fields = ["name", "workdirs", "private"]



#!/usr/bin/env python

__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2020-2021, Vanessa Sochat"
__license__ = "MPL 2.0"

from snakeface.logger import setup_logger
from django.core.wsgi import get_wsgi_application
from django.core import management
from snakeface.version import __version__
import argparse
import sys
import os

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "snakeface.settings")


def get_parser():
    parser = argparse.ArgumentParser(description="Snakeface: interface to snakemake.")

    parser.add_argument(
        "--version",
        dest="version",
        help="print the version and exit.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--noreload",
        dest="noreload",
        help="Tells Django to NOT use the auto-reloader.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--verbosity",
        dest="verbosity",
        help="Verbosity (0, 1, 2, 3).",
        choices=list(range(0, 4)),
        default=0,
    )

    parser.add_argument(
        "--workdir", dest="workdir", help="Specify the working directory.", nargs="?"
    )

    deploy_group = parser.add_argument_group("SETTINGS")
    deploy_group.add_argument(
        "--auth",
        dest="auth",
        help="Authentication type to create for the interface, defaults to token.",
        choices=["token"],
        default="token",
    )

    network_group = parser.add_argument_group("NETWORKING")
    network_group.add_argument(
        "--port", dest="port", help="Port to serve application on.", default=5000
    )

    # Logging
    logging_group = parser.add_argument_group("LOGGING")

    logging_group.add_argument(
        "--quiet",
        dest="quiet",
        help="suppress logging.",
        default=False,
        action="store_true",
    )

    logging_group.add_argument(
        "--verbose",
        dest="verbose",
        help="verbose output for logging.",
        default=False,
        action="store_true",
    )

    logging_group.add_argument(
        "--log-disable-color",
        dest="disable_color",
        default=False,
        help="Disable color for snakeface logging.",
        action="store_true",
    )

    logging_group.add_argument(
        "--log-use-threads",
        dest="use_threads",
        action="store_true",
        help="Force threads rather than processes.",
    )

    parser.add_argument(
        "repo",
        nargs="?",
        help="Repository address and destination to deploy, e.g., <source> <dest>",
    )

    parser.add_argument(
        "dest",
        nargs="?",
        help="Path to clone the repository, should not exist.",
    )

    parser.add_argument(
        "--force",
        dest="force",
        help="If the folder exists, force overwrite, meaning remove and replace.",
        default=False,
        action="store_true",
    )

    description = "subparsers for Snakeface"
    subparsers = parser.add_subparsers(
        help="snakeface actions",
        title="actions",
        description=description,
        dest="command",
    )

    # print version and exit
    subparsers.add_parser("notebook", help="run a snakeface notebook")

    return parser


def main():
    """main entrypoint for snakeface"""
    parser = get_parser()

    def help(return_code=0):
        """print help, including the software version and active client
        and exit with return code.
        """
        print("\nSnakemake Interface v%s" % __version__)
        parser.print_help()
        sys.exit(return_code)

    # If an error occurs while parsing the arguments, the interpreter will exit with value 2
    args, extra = parser.parse_known_args()

    # Show the version and exit
    if args.version:
        print(__version__)
        sys.exit(0)

    # Do we want a notebook?
    notebook = args.command == "notebook"
    if notebook:
        os.environ["SNAKEFACE_NOTEBOOK"] = "yes"
        os.putenv("SNAKEFACE_NOTEBOOK", "yes")

    # If a working directory is set
    if not args.workdir or args.workdir == ".":
        args.workdir = os.getcwd()
    if args.workdir:
        os.environ["SNAKEFACE_WORKDIR"] = args.workdir
        os.putenv("SNAKEFACE_WORKDIR", args.workdir)

    application = get_wsgi_application()

    # customize django logging
    setup_logger(
        quiet=args.quiet,
        nocolor=args.disable_color,
        debug=args.verbose,
        use_threads=args.use_threads,
    )

    # Migrations
    management.call_command("makemigrations", verbosity=args.verbosity)
    for app in ["users", "main", "base"]:
        management.call_command("makemigrations", app, verbosity=args.verbosity)
    management.call_command("migrate", verbosity=args.verbosity)

    # management.call_command("qcluster", verbosity=args.verbosity)
    management.call_command(
        "collectstatic", verbosity=args.verbosity, interactive=False
    )
    management.call_command(
        "runserver", args.port, verbosity=args.verbosity, noreload=not args.noreload
    )
    sys.exit(0)


if __name__ == "__main__":
    main()



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2020-2021, Vanessa Sochat"
__license__ = "MPL 2.0"

from django.urls import path
from . import views
from . import tasks

urlpatterns = [
    path("", views.index, name="dashboard"),
    path("workflows/run/<int:wid>/<int:uid>", tasks.run_workflow, name="run_workflow"),
    path("workflows/new/", views.new_workflow, name="new_workflow"),
    path("workflows/<int:wid>/", views.view_workflow, name="view_workflow"),
    path("workflows/command/", views.workflow_command, name="workflow_command"),
    path(
        "workflows/<int:wid>/statuses/",
        views.workflow_statuses,
        name="workflow_statuses",
    ),
    path("workflows/<int:wid>/edit/", views.edit_workflow, name="edit_workflow"),
    path(
        "workflows/<int:wid>/report/",
        views.view_workflow_report,
        name="view_workflow_report",
    ),
    path("workflows/<int:wid>/delete/", views.delete_workflow, name="delete_workflow"),
    path("workflows/<int:wid>/cancel/", views.cancel_workflow, name="cancel_workflow"),
]

app_name = "main"



from django.contrib import admin
from snakeface.apps.main.models import Workflow


class WorkflowAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "status",
        "add_date",
        "snakefile",
        "workdir",
    )

    fields = (
        "command",
        "data",
        "dag",
        "error",
        "output",
        "name",
        "snakefile",
        "snakemake_id",
        "status",
        "thread",
        "retval",
        "workdir",
        "owners",
        "contributors",
    )


admin.site.register(Workflow, WorkflowAdmin)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2020-2021, Vanessa Sochat"
__license__ = "MPL 2.0"

from snakeface.settings import cfg
from django.contrib import messages
from django.shortcuts import redirect, get_object_or_404
from snakeface.apps.main.models import Workflow
from snakeface.apps.users.models import User
from snakeface.apps.main.utils import CommandRunner, ThreadRunner

import re

# Notebook run workflow functions


def run_workflow(request, wid, uid):
    """Top level function to ensure that the user has permission to do the run,
    and we direct to the correct function (notebook or not written, another backend)
    """
    workflow = get_object_or_404(Workflow, pk=wid)
    user = get_object_or_404(User, pk=uid)
    running_notebook = cfg.NOTEBOOK or cfg.NOTEBOOK_ONLY

    # Ensure the user has permission to run the workflow
    if user not in workflow.members:
        messages.info(request, "You are not allowed to run this workflow.")

    # The workflow cannot already be running
    elif workflow.status == "RUNNING":
        messages.info(request, "This workflow is already running.")

    elif run_is_allowed(request) and running_notebook:
        workflow.reset()
        t = ThreadRunner(target=doRun, args=[workflow.id, user.id])
        t.setDaemon(True)
        t.set_workflow(workflow)
        t.start()
        workflow.thread = t.thread_id
        workflow.save()
        messages.success(request, "Workflow %s has started running." % workflow.id)
    else:
        messages.info(request, "Snakeface currently only supports notebook runs.")
    return redirect("main:view_workflow", wid=workflow.id)


# Permissions


def run_is_allowed(request):
    """Given a request, check that the run is allowed meaning:
    1. If running a notebook, we aren't over quota for jobs
    2. If not running a notebook, we aren't over user or global limits
    """
    running_notebook = cfg.NOTEBOOK or cfg.NOTEBOOK_ONLY
    running_jobs = Workflow.objects.filter(status="RUNNING").count()
    allowed = True
    if (
        running_notebook
        and cfg.MAXIMUM_NOTEBOOK_JOBS
        and running_jobs >= cfg.MAXIMUM_NOTEBOOK_JOBS
    ):
        messages.info(
            request,
            "You already have the maximum %s jobs running." % cfg.MAXIMUM_NOTEBOOK_JOBS,
        )
        allowed = False

    elif (
        not running_notebook
        and cfg.USER_WORKFLOW_RUNS_LIMIT
        >= Workflow.objects.filter(user=request.user, status="RUNNING").count()
    ):
        messages.info(
            request,
            "You are at your workflow active runs limit. Finish some and try again later.",
        )
        allowed = False

    elif not running_notebook and cfg.USER_WORKFLOW_GLOBAL_RUNS_LIMIT >= running_jobs:
        messages.info(
            request,
            "The server is at the global limit of workflow runs. Try again later.",
        )
        allowed = False

    return allowed


# Statuses


def serialize_workflow_statuses(workflow):
    """A shared helper function to serialize a list of workflow statuses into
    json.
    """
    levels = {
        "debug": "primary",
        "dag_debug": "primary",
        "info": "info",
        "warning": "warning",
        "error": "danger",
    }
    data = []
    for i, status in enumerate(workflow.workflowstatus_set.all()):
        entry = status.msg
        msg = entry.get("msg", "")
        level = levels.get(entry.get("level"), "secondary")
        badge = "<span class='badge badge-%s'>%s</span>" % (
            level,
            entry.get("level", "info"),
        )

        # If it's a traceback, format as code
        if msg and re.search("traceback|exception", msg, re.IGNORECASE):
            msg = "<code>%s</code>" % msg.replace("\n", "<br>")

        entry.update(
            {
                "order": i,
                "job": entry.get("job", ""),
                "msg": msg,
                "level": badge,
            }
        )
        data.append(entry)
    return data


def doRun(wid, uid):
    """The task to run a workflow"""
    workflow = Workflow.objects.get(pk=wid)
    user = User.objects.get(pk=uid)

    runner = CommandRunner()
    workflow.status = "RUNNING"
    workflow.save()

    # Define the function to determine cancelling the run
    def cancel_workflow(wid):
        workflow = Workflow.objects.get(pk=wid)
        return workflow.status == "CANCELLED"

    # Run the command, update when finished
    runner.run_command(
        workflow.command.split(" "),
        env={"WMS_MONITOR_TOKEN": user.token},
        cancel_func=cancel_workflow,
        cancel_func_kwargs={"wid": wid},
    )
    workflow.error = "<br>".join(runner.error)
    workflow.output = "<br>".join(runner.output)
    workflow.status = "NOTRUNNING"
    workflow.retval = runner.retval
    workflow.save()



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2020-2021, Vanessa Sochat"
__license__ = "MPL 2.0"

import os
import tempfile
import yaml
import sys

from django.core.management.utils import get_random_secret_key
from snakeface.apps.users.utils import get_username
from datetime import datetime
from importlib import import_module

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

# The snakeface global conflict contains all settings.
SETTINGS_FILE = os.path.join(BASE_DIR, "settings.yml")
if not os.path.exists(SETTINGS_FILE):
    sys.exit("Global settings file settings.yml is missing in the install directory.")


# Read in the settings file to get settings
class Settings:
    """convert a dictionary of settings (from yaml) into a class"""

    def __init__(self, dictionary):
        for key, value in dictionary.items():
            setattr(self, key, value)
        setattr(self, "UPDATED_AT", datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ"))

    def __str__(self):
        return "[snakeface-settings]"

    def __repr__(self):
        return self.__str__()

    def __iter__(self):
        for key, value in self.__dict__.items():
            yield key, value


with open(SETTINGS_FILE, "r") as fd:
    cfg = Settings(yaml.load(fd.read(), Loader=yaml.FullLoader))

# For each setting, if it's defined in the environment with SNAKEFACE_ prefix, override
for key, value in cfg:
    envar = os.getenv("SNAKEFACE_%s" % key)
    if envar:
        setattr(cfg, key, envar)

# Secret Key


def generate_secret_key(filename):
    """A helper function to write a randomly generated secret key to file"""
    key = get_random_secret_key()
    with open(filename, "w") as fd:
        fd.writelines("SECRET_KEY = '%s'" % key)


# Generate secret key if doesn't exist, and not defined in environment
SECRET_KEY = os.environ.get("SECRET_KEY")
if not SECRET_KEY:
    try:
        from .secret_key import SECRET_KEY
    except ImportError:
        SETTINGS_DIR = os.path.abspath(os.path.dirname(__file__))
        generate_secret_key(os.path.join(SETTINGS_DIR, "secret_key.py"))
        from .secret_key import SECRET_KEY


# Private only should be a boolean
cfg.PRIVATE_ONLY = cfg.PRIVATE_ONLY is not None

# Set the domain name
DOMAIN_NAME = cfg.DOMAIN_NAME
if cfg.DOMAIN_PORT:
    DOMAIN_NAME = "%s:%s" % (DOMAIN_NAME, cfg.DOMAIN_PORT)

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True if os.getenv("DEBUG") != "false" else False

# Derive list of plugins enabled from the environment
PLUGINS_LOOKUP = {
    "ldap_auth": False,
    "pam_auth": False,
    "saml_auth": False,
}
PLUGINS_ENABLED = []
using_auth_backend = False
for key, enabled in PLUGINS_LOOKUP.items():
    plugin_key = "PLUGIN_%s_ENABLED" % key.upper()
    if hasattr(cfg, plugin_key) and getattr(cfg, plugin_key) is not None:

        # Don't enable auth backends if we are using a notebook
        if cfg.NOTEBOOK_ONLY or cfg.NOTEBOOK and "AUTH" in plugin_key:
            continue

        if "AUTH" in plugin_key:
            using_auth_backend = True
        PLUGINS_ENABLED.append(key)

# Does the user want a notebook? Default to this if no auth setup
if not hasattr(cfg, "NOTEBOOK"):
    cfg.NOTEBOOK = True if not using_auth_backend else None

# If the working directory isn't defined, set to pwd
if not hasattr(cfg, "WORKDIR") or not cfg.WORKDIR:
    cfg.WORKDIR = os.getcwd()

# SECURITY WARNING: App Engine's security features ensure that it is safe to
# have ALLOWED_HOSTS = ['*'] when the app is deployed. If you deploy a Django
# app not on App Engine, make sure to set an appropriate host here.
# See https://docs.djangoproject.com/en/2.1/ref/settings/
ALLOWED_HOSTS = ["*"]

# Application definition

INSTALLED_APPS = [
    "channels",
    "snakeface.apps.base",
    "snakeface.apps.api",
    "snakeface.apps.main",
    "snakeface.apps.users",
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.humanize",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django_extensions",
    "crispy_forms",
    "social_django",
    "django_gravatar",
    "django_q",
    "rest_framework",
    "rest_framework.authtoken",
    "drf_yasg",
]


CRISPY_TEMPLATE_PACK = "bootstrap4"

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

# Do we want to enable the cache?

if cfg.ENABLE_CACHE:
    MIDDLEWARE += [
        "django.middleware.cache.UpdateCacheMiddleware",
        "django.middleware.common.CommonMiddleware",
        "django.middleware.cache.FetchFromCacheMiddleware",
    ]

    CACHE_MIDDLEWARE_ALIAS = "default"
    CACHE_MIDDLEWARE_SECONDS = 86400  # one day


# If we are using a notebook, use an in memory channel layer
if cfg.NOTEBOOK:
    CHANNEL_LAYERS = {"default": {"BACKEND": "channels.layers.InMemoryChannelLayer"}}


ROOT_URLCONF = "snakeface.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
                "snakeface.context_processors.globals",
            ],
        },
    },
]

TEMPLATES[0]["OPTIONS"]["debug"] = DEBUG
WSGI_APPLICATION = "snakeface.wsgi.application"
ASGI_APPLICATION = "snakeface.asgi.application"

AUTH_USER_MODEL = "users.User"
SOCIAL_AUTH_USER_MODEL = "users.User"
GRAVATAR_DEFAULT_IMAGE = "retro"


# Cache to tmp
CACHE_LOCATION = os.path.join(tempfile.gettempdir(), "snakeface-cache")
CACHES = {
    "default": {
        "BACKEND": "django.core.cache.backends.filebased.FileBasedCache",
        "LOCATION": CACHE_LOCATION,
    }
}

if not os.path.exists(CACHE_LOCATION):
    os.mkdir(CACHE_LOCATION)


# Database
# https://docs.djangoproject.com/en/2.1/ref/settings/#databases

# Case 1: we are running locally but want to do migration, etc. (set False to True)
if True and os.getenv("APP_ENGINE_HOST") != None:
    print("Warning: connecting to production database.")

    # Running in development, but want to access the Google Cloud SQL instance in production.
    DATABASES = {
        "default": {
            "ENGINE": "django.db.backends.postgresql",
            "USER": os.getenv("APP_ENGINE_USERNAME"),
            "PASSWORD": os.getenv("APP_ENGINE_PASSWORD"),
            "NAME": os.getenv("APP_ENGINE_DATABASE"),
            "HOST": os.getenv("APP_ENGINE_HOST"),  # Set to IP address
            "PORT": "",  # empty string for default.
        }
    }

# Case 2: we are running on app engine
elif os.getenv("APP_ENGINE_CONNECTION_NAME") != None:

    # Ensure debug is absolutely off
    TEMPLATES[0]["OPTIONS"]["debug"] = False
    DEBUG = False

    # Running on production App Engine, so connect to Google Cloud SQL using
    # the unix socket at /cloudsql/<your-cloudsql-connection string>
    DATABASES = {
        "default": {
            "ENGINE": "django.db.backends.postgresql",
            "HOST": "/cloudsql/%s" % os.getenv("APP_ENGINE_CONNECTION_NAME"),
            "USER": os.getenv("APP_ENGINE_USERNAME"),
            "PASSWORD": os.getenv("APP_ENGINE_PASSWORD"),
            "NAME": os.getenv("APP_ENGINE_DATABASE"),
        }
    }

# Case 3: Database local development uses DATABASE_* variables
elif os.getenv("DATABASE_HOST") is not None:
    # Make sure to export all of these in your .env file
    DATABASES = {
        "default": {
            "ENGINE": os.environ.get("DATABASE_ENGINE", "django.db.backends.mysql"),
            "HOST": os.environ.get("DATABASE_HOST"),
            "USER": os.environ.get("DATABASE_USER"),
            "PASSWORD": os.environ.get("DATABASE_PASSWORD"),
            "NAME": os.environ.get("DATABASE_NAME"),
        }
    }
else:
    # Use sqlite when testing locally
    DATABASES = {
        "default": {
            "ENGINE": "django.db.backends.sqlite3",
            "NAME": os.path.join(BASE_DIR, "db.sqlite3"),
        }
    }


# Password validation
# https://docs.djangoproject.com/en/2.1/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",  # noqa: 501
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",  # noqa: 501
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",  # noqa: 501
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",  # noqa: 501
    },
]

# Django Q
# workers defaults to multiprocessing CPU count, can be set if neede
# This can be sped up running with another database

Q_CLUSTER = {
    "name": "snakecluster",
    "timeout": 90,
    "retry": 120,
    "queue_limit": 50,
    "bulk": 10,
    "orm": "default",
}


# Internationalization
# https://docs.djangoproject.com/en/2.1/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# TODO: we probably want to put these in one spot relative to user home
# https://docs.djangoproject.com/en/2.1/howto/static-files/

STATIC_ROOT = "static"
STATIC_URL = "/static/"
MEDIA_ROOT = "data"
MEDIA_URL = "/data/"

# Rate Limiting

VIEW_RATE_LIMIT = "1000/1d"  # The rate limit for each view, django-ratelimit, "50 per day per ipaddress)
VIEW_RATE_LIMIT_BLOCK = (
    True  # Given that someone goes over, are they blocked for the period?
)

# On any admin or plugin login redirect to standard social-auth entry point for agreement to terms
LOGIN_REDIRECT_URL = "/login"
LOGIN_URL = "/login"

# If we are using a notebook, grab the user that started
cfg.USERNAME = None
if cfg.NOTEBOOK or cfg.NOTEBOOK_ONLY:
    cfg.USERNAME = get_username()

AUTHENTICATION_BACKENDS = ["django.contrib.auth.backends.ModelBackend"]

## PLUGINS #####################################################################

# Apply any plugin settings
for plugin in PLUGINS_ENABLED:

    plugin_module = "snakeface.plugins." + plugin
    plugin = import_module(plugin_module)

    # Add the plugin to INSTALLED APPS
    INSTALLED_APPS.append(plugin_module)

    # Add AUTHENTICATION_BACKENDS if defined, for authentication plugins
    if hasattr(plugin, "AUTHENTICATION_BACKENDS"):
        AUTHENTICATION_BACKENDS = (
            AUTHENTICATION_BACKENDS + plugin.AUTHENTICATION_BACKENDS
        )

    # Add custom context processors, if defines for plugin
    if hasattr(plugin, "CONTEXT_PROCESSORS"):
        for context_processor in plugin.CONTEXT_PROCESSORS:
            TEMPLATES[0]["OPTIONS"]["context_processors"].append(context_processor)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2020-2021, Vanessa Sochat"
__license__ = "MPL 2.0"

from django.db.models.signals import pre_save
from django.db import models

from django.conf import settings
from django.urls import reverse
from django.contrib.postgres.fields import JSONField as DjangoJSONField

from snakeface.apps.main.utils import CommandRunner, write_file, get_tmpfile, read_file
from snakeface.argparser import SnakefaceParser
from snakeface.settings import cfg
from django.db.models import Field

import itertools
import json
import os


PRIVACY_CHOICES = (
    (False, "Public (The workflow collection will be accessible by anyone)"),
    (True, "Private (The workflow collection will be not listed.)"),
)


class JSONField(DjangoJSONField):
    pass


if "sqlite" in settings.DATABASES["default"]["ENGINE"]:

    class JSONField(Field):
        def db_type(self, connection):
            return "text"

        def from_db_value(self, value, expression, connection):
            if value is not None:
                return self.to_python(value)
            return value

        def to_python(self, value):
            if value is not None:
                try:
                    return json.loads(value)
                except (TypeError, ValueError):
                    return value
            return value

        def get_prep_value(self, value):
            if value is not None:
                return str(json.dumps(value))
            return value

        def value_to_string(self, obj):
            return self.value_from_object(obj)


RUNNING_CHOICES = [
    ("RUNNING", "RUNNING"),
    ("NOTRUNNING", "NOTRUNNING"),
    ("CANCELLED", "CANCELLED"),
]


class Workflow(models.Model):
    """A workflow is associated with a specific git repository and one or more
    workflow runs.
    """

    add_date = models.DateTimeField("date published", auto_now_add=True)
    command = models.TextField(blank=False, null=False)
    data = JSONField(blank=False, null=False, default="{}")
    dag = models.TextField(blank=True, null=True)
    error = models.TextField(blank=True, null=True)
    output = models.TextField(blank=True, null=True)
    modify_date = models.DateTimeField("date modified", auto_now=True)
    name = models.CharField(max_length=250, unique=True, blank=True, null=True)
    snakefile = models.TextField(blank=False, null=False, max_length=250)
    snakemake_id = models.TextField(blank=False, null=False)
    status = models.TextField(
        choices=RUNNING_CHOICES, default="NOTRUNNING", blank=False, null=False
    )
    thread = models.PositiveIntegerField(default=None, blank=True, null=True)
    retval = models.PositiveIntegerField(default=None, blank=True, null=True)
    workdir = models.TextField(blank=False, null=False, max_length=250)

    owners = models.ManyToManyField(
        "users.User",
        blank=True,
        default=None,
        related_name="workflow_owners",
        related_query_name="owners",
    )
    contributors = models.ManyToManyField(
        "users.User",
        related_name="workflow_contributors",
        related_query_name="contributor",
        blank=True,
        help_text="users with edit permission to the workflow",
        verbose_name="Contributors",
    )

    def get_absolute_url(self):
        return_cid = self.id
        return reverse("workflow_details", args=[str(return_cid)])

    # By default, collections are public
    private = models.BooleanField(
        choices=PRIVACY_CHOICES,
        default=cfg.PRIVATE_ONLY,
        verbose_name="Accessibility",
    )

    @property
    def message_fields(self):
        fields = set()
        for status in self.workflowstatus_set.all():
            [fields.add(x) for x in status.msg]
        return fields

    def has_view_permission(self):
        if cfg.NOTEBOOK or cfg.NOTEBOOK_ONLY:
            return True
        return (self.private and self.request.user in self.members) or not self.private

    def update_command(self, command=None, do_save=False):
        """Given a command (or an automated save from the signal) update
        the command for the workflow.
        """
        if command:
            self.command = command
        else:
            parser = SnakefaceParser()
            parser.load(self.data)
            self.command = parser.command + " --wms-monitor-arg id=%s" % self.id
        if do_save:
            self.save()

    def update_dag(self, do_save=False):
        """given a snakefile, run the command to update the dag"""
        if self.snakefile and os.path.exists(self.snakefile):
            runner = CommandRunner()

            # First generate the dag, save to temporary dot file
            runner.run_command(
                ["snakemake", "--dag"], cwd=os.path.dirname(self.snakefile)
            )
            filename = write_file(
                get_tmpfile("snakeface-dag-", ".dot"), "".join(runner.output)
            )

            # Next generate the svg with dot, save to model
            runner.run_command(
                ["dot", "-Tsvg", os.path.basename(filename)],
                cwd=os.path.dirname(filename),
            )
            self.dag = "".join(runner.output)
            os.remove(filename)

            # If running from a signal, would be infinite loop
            if do_save:
                self.save()

    def __str__(self):
        return "[workflow:%s]" % self.name

    def get_label(self):
        return "workflow"

    @property
    def members(self):
        return list(itertools.chain(self.owners.all(), self.contributors.all()))

    def get_report(self):
        """load the report file, if it exists."""
        report_file = self._get_report_file()
        if report_file:
            return read_file(report_file)

    def reset(self):
        """Empty all run related fields to prepare for a new run."""
        self.output = None
        self.error = None
        self.retval = None
        self.workflowstatus_set.all().delete()
        self.save()

    def has_report(self):
        """returns True if the workflow command has a designated report, and
        the report file exists
        """
        if self._get_report_file():
            return True
        return False

    def _get_report_file(self):
        report_file = self.data.get("report")
        fullpath = None
        if report_file:
            fullpath = os.path.join(self.workdir, report_file)
        return fullpath

    def has_edit_permission(self):
        """If we are running in a notebook environment, there is just one user
        that has edit access to anything. Otherwise, the user must be an owner
        """
        if cfg.NOTEBOOK or cfg.NOTEBOOK_ONLY:
            return True
        return (
            not self.request.user.is_anonymous
            and self.request.user in self.members
            and self.private
        )

    class Meta:
        app_label = "main"


class WorkflowStatus(models.Model):
    """A workflow status is a status message send from running a workflow"""

    # executor = models.TextField(null=False, blank=False)
    add_date = models.DateTimeField("date published", auto_now_add=True)
    modify_date = models.DateTimeField("date modified", auto_now=True)
    msg = JSONField(blank=False, null=False, default="{}")
    workflow = models.ForeignKey(
        "main.Workflow", null=False, blank=False, on_delete=models.CASCADE
    )


def update_workflow(sender, instance, **kwargs):
    instance.update_dag()
    instance.update_command()


pre_save.connect(update_workflow, sender=Workflow)



from django import template

register = template.Library()


@register.filter(name="replace", is_safe=True)
def replace(string, original, value):
    return string.replace(original, value)


@register.filter
def lookup(d, key):
    return d[key]


@register.filter
def range(min=5):
    return range(min)


@register.filter
def index(indexable, i):
    return indexable[i]



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2020, Vanessa Sochat"
__license__ = "MPL 2.0"

from django.http import HttpResponse

# Warmup requests for app engine


def warmup():
    return HttpResponse(status=200)



from django.urls import re_path

from . import consumers

websocket_urlpatterns = [
    re_path(
        r"ws/workflows/(?P<wid>\w+)/$",
        consumers.WorkflowConsumer.as_asgi(),
        name="workflow_status_websocket",
    ),
]

app_name = "main"



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2020-2021, Vanessa Sochat"
__license__ = "MPL 2.0"

from django.contrib.sites.shortcuts import get_current_site
from snakeface import settings


def globals(request):
    """Returns a dict of defaults to be used by templates, if configured
    correcty in the settings.py file."""
    return {
        "DOMAIN": settings.DOMAIN_NAME,
        "WORKFLOW_UPDATE_SECONDS": settings.cfg.WORKFLOW_UPDATE_SECONDS,
        "NOTEBOOK": settings.cfg.NOTEBOOK,
        "TWITTER_USERNAME": settings.cfg.TWITTER_USERNAME,
        "GITHUB_REPOSITORY": settings.cfg.GITHUB_REPOSITORY,
        "GITHUB_DOCUMENTATION": settings.cfg.GITHUB_DOCUMENTATION,
        "SITE_NAME": get_current_site(request).name,
        "GOOGLE_ANALYTICS_ID": settings.cfg.GOOGLE_ANALYTICS_ID,
        "GOOGLE_ANALYTICS_SITE": settings.cfg.GOOGLE_ANALYTICS_SITE,
    }



from django.conf.urls import url
from django.urls import path

from drf_yasg.views import get_schema_view
from drf_yasg import openapi

from snakeface.settings import cfg
import snakeface.apps.api.views as api_views
from .permissions import AllowAnyGet

# Documentation URL
schema_view = get_schema_view(
    openapi.Info(
        title="SnakeFace API",
        default_version="v1",
        description="API for Snakemake to interact with Snakeface",
        license=openapi.License(name="Apache License"),
        contact=openapi.Contact(url=cfg.HELP_CONTACT_URL),
    ),
    public=True,
    permission_classes=(AllowAnyGet,),
)

urlpatterns = [
    url(
        r"^swagger(?P<format>\.json|\.yaml)$",
        schema_view.without_ui(cache_timeout=0),
        name="schema-json",
    ),
    url(
        r"^swagger/$",
        schema_view.with_ui("swagger", cache_timeout=0),
        name="schema-swagger-ui",
    ),
    url(
        r"^redoc/$", schema_view.with_ui("redoc", cache_timeout=0), name="schema-redoc"
    ),
    path(
        "api/service-info",
        api_views.ServiceInfo.as_view(),
        name="service_info",
    ),
    path(
        "create_workflow",
        api_views.CreateWorkflow.as_view(),
        name="create_workflow",
    ),
    path(
        "update_workflow_status",
        api_views.UpdateWorkflow.as_view(),
        name="update_workflow_status",
    ),
]


app_name = "api"



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2020, Vanessa Sochat"
__license__ = "MPL 2.0"

from django.conf.urls import url, include
from snakeface.apps.users import views
from social_django import urls as social_urls

urlpatterns = [
    #    url(r"^login/$", views.login, name="login"),
    url(r"^login/notebook/$", views.notebook_login, name="notebook_login"),
    # url(r"^accounts/login/$", views.login),
    url(r"^logout/$", views.logout, name="logout"),
    url("", include(social_urls, namespace="social")),
]

app_name = "users"



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2020-2021, Vanessa Sochat"
__license__ = "MPL 2.0"

from snakemake import get_argument_parser
from snakeface.settings import cfg
from snakeface import settings
from snakeface.apps.main.utils import get_snakefile_choices
from jinja2 import Template
import logging
import os
import json

logger = logging.getLogger("argparser")

# Prepare path to templates
here = os.path.abspath(os.path.dirname(__file__))
templates = os.path.join(here, "apps", "main", "templates", "forms")


class SnakefaceParser:
    """A Snakeface Parser is a wrapper to an argparse.Parser, and aims
    to make it easy to loop over arguments and options, and generate various
    representations (e.g., an input field) for the interface. The point is
    not to use it to parse arguments and validate, but to output all
    fields to a front end form.
    """

    # Update the listing of snakefiles on the parser init
    snakefiles = get_snakefile_choices()

    def __init__(self):
        """load the parser, optionally specifying a profile"""
        self.parser = get_argument_parser()
        self._groups = {}
        self._args = {}
        self.groups
        self._errors = []

        # A profile can further customize job submission
        if cfg.PROFILE and os.path.exists(cfg.PROFILE):
            print("Loading profile %s" % cfg.PROFILE)
            self.parser = get_argument_parser(cfg.PROFILE)

    def __str__(self):
        return "[snakeface-parser]"

    def __repr__(self):
        return self.__str__()

    @property
    def errors(self):
        return " ".join(self._errors)

    def get(self, name, default=None):
        """A general get function to return an argument that might be nested under
        a group. These objects are the same as linked in _groups.
        """
        return self._args.get(name, default)

    def load(self, argdict):
        """Load is a wrapper around set - we loop through a dictionary and set all
        arguments.
        """
        if isinstance(argdict, str):
            argdict = json.loads(argdict)
        for key, value in argdict.items():
            arg = self._args.get(key)
            if arg:
                arg.value = value

    def set(self, name, value):
        """Set a value for an argument. This is typically what the user has selected."""
        arg = self._args.get(name)
        if arg:
            arg.value = value

    def to_dict(self):
        """the opposite of load, this function exports an argument"""
        return {name: arg.value for name, arg in self._args.items()}

    @property
    def snakefile(self):
        snakefile = self._args.get("snakefile")
        if snakefile:
            return snakefile.value

    def validate(self):
        """ensure that all required args are defined"""
        valid = True
        for key in self.required:
            if not self._args.get(key):
                self._errors.append("The %s is required." % key)
                valid = False
        return valid

    @property
    def required(self):
        return ["cores", "snakefile"]

    @property
    def command(self):
        """Given a loaded set of arguments, generate the command."""
        command = "snakemake"

        for name, arg in self._args.items():
            if arg.value:

                # If the value is set to the default, ignore it
                if arg.value == arg.action["default"] and name not in self.required:
                    continue

                flag = ""
                if arg.action["option_strings"]:
                    flag = arg.action["option_strings"][0]

                # Assemble argument based on type
                if arg.is_boolean:
                    command += " %s" % flag
                else:
                    command += " %s %s" % (flag, arg.value)

        return command

    @property
    def groups(self):
        """yield arguments organized by groups, with the intention to easily map
        into a form on the front end. The groups seem to have ALL arguments each,
        so we have to artificially separate them.
        """
        if self._groups:
            return self._groups

        # Generate an argument lookup based on dest
        lookup = {}
        for action in self.parser._actions:
            lookup[action.dest] = SnakefaceArgument(
                action, action.dest in self.required
            )

            # Define choices
            if action.dest == "snakefile":
                lookup[action.dest].update_choice_fields({"snakefile": self.snakefiles})

            # Set the wms monitor to be this server
            if action.dest == "wms_monitor":
                lookup[action.dest].value = settings.DOMAIN_NAME

        # This top level organizes into groups
        for group in self.parser._action_groups:
            group_dict = {
                a.dest: lookup.get(a.dest)
                for a in group._group_actions
                if self.include_argument(a.dest, group.title)
            }

            # Store a flattened representation to manipulate later
            self._args.update(group_dict)

            # Don't add empty groups
            if group_dict:
                self._groups[group.title] = group_dict
        return self._groups

    def include_argument(self, name, group):
        """Given an argument name, and a group name, skip if settings disable
        it
        """
        # Never include these named arguments
        if name in ["help", "version"]:
            return False

        # Skip groups based on specific configuration settings
        if not cfg.EXECUTOR_CLUSTER and group == "CLUSTER":
            return False
        if not cfg.EXECUTOR_GOOGLE_LIFE_SCIENCES and group == "GOOGLE_LIFE_SCIENCE":
            return False
        if not cfg.EXECUTOR_KUBERNETES and group == "KUBERNETES":
            return False
        if not cfg.EXECUTOR_TIBANNA and group == "TIBANNA":
            return False
        if not cfg.EXECUTOR_TIBANNA and group == "TIBANNA":
            return False
        if not cfg.EXECUTOR_GA4GH_TES and group == "TES":
            return False
        if cfg.DISABLE_SINGULARITY and group == "SINGULARITY":
            return False
        if cfg.DISABLE_CONDA and group == "CONDA":
            return False
        if cfg.DISABLE_NOTEBOOKS and group == "NOTEBOOKS":
            return False
        return True


class SnakefaceArgument:
    """A Snakeface argument takes an action from a parser, and is able to
    easily generate front end views (e.g., a form element) for it
    """

    def __init__(self, action, required=False):
        self.action = action.__dict__
        self.boolean_template = ""
        self.text_template = ""
        self.choice_template = ""
        self.choice_fields = {}
        self.value = ""
        self.required = required

    def __str__(self):
        return self.action["dest"]

    def __repr__(self):
        return self.__str__()

    def update_choice_fields(self, updates):
        self.choice_fields.update(updates)

    @property
    def field_name(self):
        return " ".join([x.capitalize() for x in self.action["dest"].split("_")])

    @property
    def is_boolean(self):
        return self.action["nargs"] == 0 and self.action["const"]

    def field(self):
        """generate a form field for the argument"""
        if self.action["dest"] in self.choice_fields:
            return self.choice_field()
        if self.is_boolean:
            return self.boolean_field()
        return self.text_field()

    def load_template(self, path):
        """Given a path to a template file, load the template with jinja2"""
        if os.path.exists(path):
            with open(path, "r") as fd:
                template = Template(fd.read())
            return template
        logging.warning("%s does not exist, no template loaded.")
        return ""

    def boolean_field(self):
        """generate a boolean field (radio button) via a jinja2 template"""
        # Ensure that we only load/read the file once
        if not self.boolean_template:
            self.boolean_template = self.load_template(
                os.path.join(templates, "boolean_field.html")
            )
        checked = "checked" if self.action["default"] == True else ""
        return self.boolean_template.render(
            label=self.field_name,
            help=self.action["help"],
            name=self.action["dest"],
            checked=checked,
            required="required" if self.required else "",
        )

    def text_field(self):
        """generate a text field for using a pre-loaded jinja2 template"""
        if not self.text_template:
            self.text_template = self.load_template(
                os.path.join(templates, "text_field.html")
            )

        return self.text_template.render(
            name=self.action["dest"],
            default=self.action["default"] or self.value,
            label=self.field_name,
            help=self.action["help"],
            required="required" if self.required else "",
        )

    def choice_field(self):
        """generate a choice field for using a pre-loaded jinja2 template"""
        if not self.choice_template:
            self.choice_template = self.load_template(
                os.path.join(templates, "choice_field.html")
            )

        return self.choice_template.render(
            name=self.action["dest"],
            label=self.field_name,
            help=self.action["help"],
            required="required" if self.required else "",
            choices=self.choice_fields.get(self.action["dest"]),
        )



import json
import asyncio
from channels.generic.websocket import AsyncJsonWebsocketConsumer
from snakeface.apps.main.models import Workflow
from snakeface.apps.main.tasks import serialize_workflow_statuses
from snakeface.settings import cfg
from asgiref.sync import sync_to_async


@sync_to_async
def get_statuses(workflow_id):
    """Return a dictionary of workflow statuses on success. If the workflow
    doesn't exist, then return False and we disconnect from the socket.
    """
    try:
        workflow = Workflow.objects.get(id=workflow_id)
        return {
            "statuses": serialize_workflow_statuses(workflow),
            "output": workflow.output,
            "error": workflow.error,
            "retval": workflow.retval,
        }
    except:
        return False


async_get_statuses = sync_to_async(get_statuses, thread_sensitive=True)


class WorkflowConsumer(AsyncJsonWebsocketConsumer):
    async def connect(self):
        self.workflow_id = self.scope["path"].strip("/").split("/")[-1]
        print("websocket connect for workflow %s" % self.workflow_id)
        self.connected = True
        await self.channel_layer.group_add(self.workflow_id, self.channel_name)
        await self.accept()
        asyncio.create_task(self.update_workflow_status())

    async def update_workflow_status(self):
        while self.connected:
            await asyncio.sleep(cfg.WORKFLOW_UPDATE_SECONDS)

            status = "success"
            data = await async_get_statuses(self.workflow_id)
            if data == False:
                data = {
                    "message": "Workflow with id %s does not exist." % self.workflow_id
                }
                status = "error"
                self.connected = False

            await self.send_json(
                {"type": "websocket.send", "text": data, "status": status}
            )

    async def disconnect(self, close_code):
        self.connected = False
        await self.channel_layer.group_discard(self.workflow_id, self.channel_name)

    async def receive(self, text_data):
        print("receive", text_data)
        text_data_json = json.loads(text_data)
        message = text_data_json.get("message")

        # Send message to room group
        await self.channel_layer.group_send(
            self.room_group_name, {"type": "chat_message", "message": message}
        )



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2020-2021, Vanessa Sochat"
__license__ = "MPL 2.0"

from django.shortcuts import render, redirect, get_object_or_404
from django.http import HttpResponseForbidden, JsonResponse
from django.contrib import messages
import os

from ratelimit.decorators import ratelimit
from snakeface.argparser import SnakefaceParser
from snakeface.settings import cfg
from snakeface.apps.main.models import Workflow
from snakeface.apps.main.forms import WorkflowForm
from snakeface.apps.main.tasks import run_workflow, serialize_workflow_statuses
from snakeface.apps.users.decorators import login_is_required
from snakeface.settings import (
    VIEW_RATE_LIMIT as rl_rate,
    VIEW_RATE_LIMIT_BLOCK as rl_block,
)


# Dashboard


@login_is_required
@ratelimit(key="ip", rate=rl_rate, block=rl_block)
def index(request):
    workflows = None
    if request.user.is_authenticated:
        workflows = Workflow.objects.filter(owners=request.user)
    return render(
        request,
        "main/index.html",
        {"workflows": workflows, "page_title": "Dashboard"},
    )


# Workflows


@login_is_required
@ratelimit(key="ip", rate=rl_rate, block=rl_block)
def delete_workflow(request, wid):
    workflow = get_object_or_404(Workflow, pk=wid)

    # Ensure that the user is an owner
    if request.user not in workflow.owners.all():
        return HttpResponseForbidden()
    workflow.delete()
    return redirect("main:dashboard")


@login_is_required
@ratelimit(key="ip", rate=rl_rate, block=rl_block)
def cancel_workflow(request, wid):
    workflow = get_object_or_404(Workflow, pk=wid)

    # Ensure that the user is an owner
    if request.user not in workflow.owners.all():
        return HttpResponseForbidden()
    workflow.status = "CANCELLED"
    workflow.save()
    messages.info(
        request, "Your workflow has been cancelled, and will stop within 10 seconds."
    )
    return redirect("main:view_workflow", wid=workflow.id)


@login_is_required
@ratelimit(key="ip", rate=rl_rate, block=rl_block)
def edit_workflow(request, wid):

    workflow = get_object_or_404(Workflow, pk=wid)

    # Ensure that the user is an owner
    if request.user not in workflow.owners.all():
        return HttpResponseForbidden()

    # Give a warning if the snakefile doesn't exist
    if not os.path.exists(workflow.snakefile):
        messages.warning(
            request, "Warning: This snakefile doesn't appear to exist anymore."
        )

    # Create and update a parser with the current settings
    parser = SnakefaceParser()
    parser.load(workflow.data)
    return edit_or_update_workflow(request, workflow=workflow, parser=parser)


@login_is_required
@ratelimit(key="ip", rate=rl_rate, block=rl_block)
def new_workflow(request):
    parser = SnakefaceParser()
    if request.user.is_authenticated:
        return edit_or_update_workflow(request, parser=parser)
    return HttpResponseForbidden()


def edit_or_update_workflow(request, parser, workflow=None):
    """A shared function to edit or update an existing workflow."""

    # Ensure the user has permission to update
    if workflow:
        existed = True
        action = "update"
        if request.user not in workflow.owners.all():
            return HttpResponseForbidden()
    else:
        workflow = Workflow()
        action = "create"
        existed = False

    form = WorkflowForm(request.POST or None, instance=workflow)

    # Case 1: parse a provided form to update current data
    if request.method == "POST" and form.is_valid():

        for arg, setting in request.POST.items():
            parser.set(arg, setting)

        # Has the user gone over the workflow number limit?
        if (
            Workflow.objects.filter(owners=request.user).count()
            >= cfg.USER_WORKFLOW_LIMIT
        ):
            messages.info(
                request, "You are at the workflow limit of %s" % cfg.USER_WORKFLOW_LIMIT
            )
        elif not parser.validate():
            messages.info(request, parser.errors)
        else:
            print("Creating workflow")
            workflow = form.save()
            workflow.data = parser.to_dict()
            workflow.snakefile = parser.snakefile
            workflow.workdir = request.POST.get("workdirs", cfg.WORKDIR)
            workflow.private = (
                True if cfg.PRIVATE_ONLY else request.POST.get("private", 1) == 1
            )
            workflow.owners.add(request.user)
            # Save updates the dag and command
            workflow.save()
            return run_workflow(request=request, wid=workflow.id, uid=request.user.id)

    # Case 2: no snakefiles:
    if not parser.snakefiles:
        message = (
            "No Snakefiles were found in any path under %s."
            " You must have one to %s a workflow." % (cfg.WORKDIR, action)
        )
        messages.info(request, message)
        return redirect("main:dashboard")

    # Case 3: Render an empty form with current working directory
    if existed:
        form.fields["workdirs"].initial = workflow.workdir
    return render(
        request,
        "workflows/new.html",
        {
            "groups": parser.groups,
            "page_title": "%s Workflow" % action.capitalize(),
            "form": form,
            "workflow_id": getattr(workflow, "id", None),
        },
    )


@login_is_required
@ratelimit(key="ip", rate=rl_rate, block=rl_block)
def workflow_command(request):
    """is called from the browser via POST to update the command"""
    parser = SnakefaceParser()
    if request.method == "POST":
        for arg, setting in request.POST.items():
            parser.set(arg, setting)
        return JsonResponse({"command": parser.command})


@login_is_required
def workflow_statuses(request, wid):
    """return serialized workflow statuses for the details view."""
    workflow = get_object_or_404(Workflow, pk=wid)
    return JsonResponse({"data": serialize_workflow_statuses(workflow)})


@login_is_required
@ratelimit(key="ip", rate=rl_rate, block=rl_block)
def view_workflow(request, wid):

    workflow = get_object_or_404(Workflow, pk=wid)
    return render(
        request,
        "workflows/detail.html",
        {
            "workflow": workflow,
            "page_title": "%s: %s" % (workflow.name or "Workflow", workflow.id),
        },
    )


def view_workflow_report(request, wid):
    """If a workflow generated a report and the report exists, render it to a page"""
    workflow = get_object_or_404(Workflow, pk=wid)
    report = workflow.get_report()
    if not report:
        messages.info(request, "This workflow does not have a report file.")
        redirect("main:view_workflow", wid=workflow.id)
    return render(
        request,
        "workflows/report.html",
        {"workflow": workflow, "page_title": "Report", "report": report},
    )



from django.conf import settings

from rest_framework.renderers import JSONRenderer
from ratelimit.mixins import RatelimitMixin
from django.shortcuts import get_object_or_404

from snakeface.apps.main.models import Workflow, WorkflowStatus
from snakeface.settings import cfg
from snakeface.version import __version__
from rest_framework.response import Response
from rest_framework.views import APIView
from .permissions import check_user_authentication

import json


class ServiceInfo(RatelimitMixin, APIView):
    """Return a 200 response to indicate a running service. Note that we are
    not currently including all required fields. See:
    https://ga4gh.github.io/workflow-execution-service-schemas/docs/#operation/GetServiceInfo
    """

    ratelimit_key = "ip"
    ratelimit_rate = settings.VIEW_RATE_LIMIT
    ratelimit_block = settings.VIEW_RATE_LIMIT_BLOCK
    ratelimit_method = "GET"
    renderer_classes = (JSONRenderer,)

    def get(self, request):
        print("GET /service-info")

        data = {
            "id": "snakeface",
            "status": "running",  # Extra field looked for by Snakemake
            "name": "Snakemake Workflow Interface (SnakeFace)",
            "type": {"group": "org.ga4gh", "artifact": "beacon", "version": "1.0.0"},
            "description": "This service provides an interface to interact with Snakemake.",
            "organization": {"name": "Snakemake", "url": "https://snakemake.github.io"},
            "contactUrl": cfg.HELP_CONTACT_URL,
            "documentationUrl": "https://snakemake.github.io/snakeface",
            "createdAt": "2020-12-04T12:57:19Z",
            "updatedAt": cfg.UPDATED_AT,
            "environment": cfg.ENVIRONMENT,
            "version": __version__,
            "auth_instructions_url": "",
        }

        # Must make model json serializable
        return Response(status=200, data=data)


class CreateWorkflow(RatelimitMixin, APIView):
    """Create a snakemake workflow. Given that we provide an API token, we
    expect the workflow model to already be created and simply generate a run
    for it.
    """

    ratelimit_key = "ip"
    ratelimit_rate = settings.VIEW_RATE_LIMIT
    ratelimit_block = settings.VIEW_RATE_LIMIT_BLOCK
    ratelimit_method = "GET"
    renderer_classes = (JSONRenderer,)

    def get(self, request):
        print("GET /create_workflow")

        # If the request provides an id, check for workflow
        workflow = request.GET.get("id")
        user = None

        if workflow:
            workflow = get_object_or_404(Workflow, pk=workflow)

        # Does the server require authentication?
        if cfg.REQUIRE_AUTH:
            user, response_code = check_user_authentication(request)
            if not user:
                return Response(status=response_code)

            # If we have a workflow, check that user has permission to use/update
            if workflow and user not in workflow.owners.all():
                return Response(status=403)

        # If we don't have a workflow, create one
        if workflow:

            # Remove old statuses here
            workflow.workflowstatus_set.all().delete()

        else:
            # Add additional metadata to creation
            snakefile = request.POST.get("snakefile")
            workdir = request.POST.get("workdir")
            command = request.POST.get("command")
            workflow = Workflow(snakefile=snakefile, workdir=workdir, command=command)
            workflow.save()
            if user:
                workflow.owners.add(user)

        data = {"id": workflow.id}
        return Response(status=200, data=data)


class UpdateWorkflow(RatelimitMixin, APIView):
    """Update an existing snakemake workflow. Authentication is required,
    and the workflow must exist.
    """

    ratelimit_key = "ip"
    ratelimit_rate = settings.VIEW_RATE_LIMIT
    ratelimit_block = settings.VIEW_RATE_LIMIT_BLOCK
    ratelimit_method = "POST"
    renderer_classes = (JSONRenderer,)

    def post(self, request):
        print("POST /update_workflow_status")

        # We must have an existing workflow to update
        workflow = get_object_or_404(Workflow, pk=request.POST.get("id"))

        # Does the server require authentication?
        if cfg.REQUIRE_AUTH:
            user, response_code = check_user_authentication(request)
            if not user:
                return Response(response_code)

            # If we have a workflow, check that user has permission to use/update
            if workflow and user not in workflow.owners.all():
                return Response(403)

        # The message should be json dump of attributes
        message = json.loads(request.POST.get("msg", {}))

        # Update the workflow with a new status message
        WorkflowStatus.objects.create(workflow=workflow, msg=message)
        return Response(status=200, data={})



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2020, Vanessa Sochat"
__license__ = "MPL 2.0"

from django.views.generic.base import TemplateView
from django.urls import path
from django.conf.urls import url

from . import views

urlpatterns = [
    path("_ah/warmup/", views.warmup, name="warmup"),
    url(
        r"^robots\.txt/$",
        TemplateView.as_view(
            template_name="base/robots.txt", content_type="text/plain"
        ),
    ),
]

app_name = "base"
